package com.raritan.json_rpc;

import java.util.HashMap;
import java.util.Map;

import org.json.JSONException;
import org.json.JSONObject;

import com.raritan.idl.TypeInfo;

public abstract class ValueObjectCodec {

    /* methods to be implemented by value object codec classes */
    public abstract void encodeValObj(JSONObject json, ValueObject vo) throws JSONException;

    public abstract ValueObject decodeValObj(ValueObject vo, JSONObject json, Agent agent) throws JSONException;

    /* static value object codec registry */
    /* package */ static Map<String, Map<TypeInfo, ValueObjectCodec>> codecMap =
	    new HashMap<String, Map<TypeInfo, ValueObjectCodec>>();

    public static void registerCodec(TypeInfo ti, ValueObjectCodec codec) {
	Map<TypeInfo, ValueObjectCodec> map = codecMap.get(ti.getTypeName());
	if (map == null) {
	    map = new HashMap<TypeInfo, ValueObjectCodec>();
	    codecMap.put(ti.getTypeName(), map);
	}
	assert(!map.containsKey(ti));
	map.put(ti, codec);
    }

    public static TypeInfo lookupFullTypeInfo(TypeInfo ti) {
	Map<TypeInfo, ValueObjectCodec> map = codecMap.get(ti.getTypeName());
	if (map == null) {
	    return null;
	}

	for (TypeInfo info : map.keySet()) {
	    if (info.equals(ti)) {
		return info;
	    }
	}
	return null;
    }

    public static JSONObject encode(final ValueObject vo) throws JSONException {
	if (vo == null) return null;

	TypeInfo ti = vo.getTypeInfo();
	Map<TypeInfo, ValueObjectCodec> map = codecMap.get(ti.getTypeName());
	ValueObjectCodec voc = map == null ? null : map.get(ti);
	if (voc == null) {
	    throw new IllegalArgumentException("No codec for " + ti);
	}

	JSONObject value = new JSONObject();
	voc.encodeValObj(value, vo);

	JSONObject json = new JSONObject();
	json.put("type", ti);
	json.put("value", value);
	return json;
    }

    public static ValueObject decodeAs(final JSONObject json, final Agent agent, final TypeInfo type) throws JSONException {
	if (json == null) {
	    return null;
	}

	TypeInfo ti = new TypeInfo(json.getString("type"), null);
	Map<TypeInfo, ValueObjectCodec> map = codecMap.get(ti.getTypeName());
	TypeInfo bestTypeInfo = null;
	if (map != null) {
	    for (Map.Entry<TypeInfo, ValueObjectCodec> entry : map.entrySet()) {
		if (ti.isCompatible(entry.getKey())) {
		    bestTypeInfo = entry.getKey();
		}
	    }
	}
	if (bestTypeInfo == null) {
	    throw new IllegalArgumentException("No codec for " + ti);
	}

	TypeInfo t = bestTypeInfo;
	while (t != null && !t.isCompatible(type)) {
	    t = t.getBase();
	}
	if (t == null) {
	    throw new IllegalArgumentException("Codec found for " + bestTypeInfo + ", but incompatible with " + type);
	}

	map = codecMap.get(type.getTypeName());
	ValueObjectCodec voc = map == null ? null : map.get(type);
	if (voc == null) {
	    throw new IllegalArgumentException("No codec for desired type " + type);
	}

	JSONObject data = json.getJSONObject("value");
	ValueObject result = voc.decodeValObj(null, data, agent);
	if (result != null) {
	    result.origData = data;
	    result.origType = bestTypeInfo;
	    result.agent = agent;
	}
	return result;
    }
}
